"""
Sum of diagonal and low rank matrices
"""
from typing import List, Iterable
import numpy as np
from numpy import ndarray
from scipy.linalg import block_diag
from spmat import utils, linalg


class ILMat:
    """
    Identity plus outer product of low rank matrix, I + L @ L.T

    Attributes
    ----------
    lmat : ndarray
        Low rank matrix.
    tmat : ILMat
        Alternative ILMat, construct by transpose of ``lmat``. It is not
        ``None`` when ``lmat`` is a 'fat' matrix.
    dsize : int
        Size of the diagonal.
    lrank : int
        Rank of the low rank matrix.

    Methods
    -------
    dot(x)
        Dot product with vector or matrix.
    invdot(x)
        Inverse dot product with vector or matrix.
    logdet()
        Log determinant of the matrix.
    """

    def __init__(self, lmat: Iterable):
        """
        Parameters
        ----------
        lmat : Iterable

        Raises
        ------
        ValueError
            When ``lmat`` is not a matrix.
        """
        self.lmat = utils.to_numpy(lmat, ndim=(2,))
        self.dsize = self.lmat.shape[0]
        self.lrank = min(self.lmat.shape)

        self._u, s, _ = np.linalg.svd(self.lmat, full_matrices=False)
        self._v = s**2
        self._w = -self._v/(1 + self._v)

    @property
    def mat(self) -> ndarray:
        return np.identity(self.dsize) + (self._u*self._v) @ self._u.T

    @property
    def invmat(self) -> ndarray:
        return np.identity(self.dsize) + (self._u*self._w) @ self._u.T

    def dot(self, x: Iterable) -> ndarray:
        """
        Dot product with vector or matrix

        Parameters
        ----------
        x : Iterable
            Vector or matrix

        Returns
        -------
        ndarray
        """
        x = utils.to_numpy(x, ndim=(1, 2))
        return x + (self._u*self._v) @ (self._u.T @ x)

    def invdot(self, x: Iterable) -> ndarray:
        """
        Inverse dot product with vector or matrix

        Parameters
        ----------
        x : Iterable
            Vector or matrix

        Returns
        -------
        ndarray
        """
        x = utils.to_numpy(x, ndim=(1, 2))
        return x + (self._u*self._w) @ (self._u.T @ x)

    def logdet(self) -> float:
        """
        Log determinant

        Returns
        -------
        float
            Log determinant of the matrix.
        """
        return np.log(1 + self._v).sum()

    def __repr__(self) -> str:
        return f"ILMat(dsize={self.dsize}, lrank={self.lrank})"


class BILMat:
    """
    Block ILMat.
    """

    def __init__(self, lmats: Iterable, dsizes: Iterable):
        self.lmats = np.ascontiguousarray(lmats)
        self.dsizes = np.asarray(dsizes).astype(int)
        self.lranks = np.minimum(self.dsizes, self.lmats.shape[1])
        self.dsize = self.dsizes.sum()

        if self.dsizes.sum() != self.lmats.shape[0]:
            raise ValueError("Sizes of blocks do not match shape of matrix.")

        self._u, s = linalg.block_lsvd(self.lmats.copy(),
                                       self.dsizes,
                                       self.lranks)
        self._v = s**2
        self._w = -self._v/(1 + self._v)

    @property
    def lmat_blocks(self) -> List[ndarray]:
        return np.split(self.lmats, np.cumsum(self.dsizes)[:-1], axis=0)

    @property
    def mat(self) -> ndarray:
        return block_diag(*[
            np.identity(self.dsizes[i]) + lmat.dot(lmat.T)
            for i, lmat in enumerate(self.lmat_blocks)
        ])

    @property
    def invmat(self) -> ndarray:
        return block_diag(*[
            np.linalg.inv(np.identity(self.dsizes[i]) + lmat.dot(lmat.T))
            for i, lmat in enumerate(self.lmat_blocks)
        ])

    def dot(self, x: Iterable) -> ndarray:
        x = np.ascontiguousarray(x)
        dotfun = linalg.block_mvdot if x.ndim == 1 else linalg.block_mmdot
        return dotfun(self._u, self._v, x, self.dsizes, self.lranks)

    def invdot(self, x: Iterable) -> ndarray:
        x = np.ascontiguousarray(x)
        dotfun = linalg.block_mvdot if x.ndim == 1 else linalg.block_mmdot
        return dotfun(self._u, self._w, x, self.dsizes, self.lranks)

    def logdet(self) -> float:
        return np.log(1 + self._v).sum()

    def __repr__(self) -> str:
        return f"BILMat(dsize={self.dsize}, num_blocks={self.dsizes.size})"


class DLMat:
    """
    Diagonal plus outer product of low rank matrix, D + L @ L.T

    Attributes
    ----------
    diag : ndarray
        Diagonal vector.
    lmat : ndarray
        Low rank matrix.
    dsize : int
        Size of the diagonal.
    lrank : int
        Rank of the low rank matrix.
    sdiag : ndarray
        Square root of diagonal vector.
    ilmat : ILMat
        Inner ILMat after strip off the diagonal vector.

    Methods
    -------
    dot(x)
        Dot product with vector or matrix.
    invdot(x)
        Inverse dot product with vector or matrix.
    logdet()
        Log determinant of the matrix.
    """

    def __init__(self, diag: Iterable, lmat: Iterable):
        """
        Parameters
        ----------
        diag : Iterable
            Diagonal vector.
        lmat : Iterable
            Low rank matrix.

        Raises
        ------
        ValueError
            If length of ``diag`` not match with number of rows of ``lmat``.
        ValueError
            If there are non-positive numbers in ``diag``.
        """
        diag = utils.to_numpy(diag, ndim=(1,))
        lmat = utils.to_numpy(lmat, ndim=(2,))
        if diag.size != lmat.shape[0]:
            raise ValueError("`diag` and `lmat` size not match.")
        if any(diag <= 0.0):
            raise ValueError("`diag` must be all positive.")

        self.diag = diag
        self.lmat = lmat

        self.dsize = self.diag.size
        self.lrank = min(self.lmat.shape)

        self.sdiag = np.sqrt(self.diag)
        self.ilmat = ILMat(self.lmat/self.sdiag[:, np.newaxis])

    @property
    def mat(self) -> ndarray:
        return np.diag(self.diag) + self.lmat.dot(self.lmat.T)

    @property
    def invmat(self) -> ndarray:
        return self.ilmat.invmat/(self.sdiag[:, np.newaxis] * self.sdiag)

    def dot(self, x: Iterable) -> ndarray:
        """
        Inverse dot product with vector or matrix

        Parameters
        ----------
        x : Iterable
            Vector or matrix

        Returns
        -------
        ndarray
        """
        x = utils.to_numpy(x, ndim=(1, 2))
        x = (x.T*self.sdiag).T
        x = self.ilmat.dot(x)
        x = (x.T*self.sdiag).T
        return x

    def invdot(self, x: Iterable) -> ndarray:
        """
        Inverse dot product with vector or matrix

        Parameters
        ----------
        x : Iterable
            Vector or matrix

        Returns
        -------
        ndarray
        """
        x = utils.to_numpy(x, ndim=(1, 2))
        x = (x.T/self.sdiag).T
        x = self.ilmat.invdot(x)
        x = (x.T/self.sdiag).T
        return x

    def logdet(self) -> float:
        """
        Log determinant

        Returns
        -------
        float
            Log determinant of the matrix.
        """
        return np.log(self.diag).sum() + self.ilmat.logdet()

    def __repr__(self) -> str:
        return f"DLMat(dsize={self.dsize}, lrank={self.lrank})"


class BDLMat:
    """
    Block diagonal low rank matrix, [... D + L @ L.T ...]

    Attributes
    ----------
    diags : ndarray
        Diagonal component of the matrix.
    lmats : ndarray
        L-Matrix component of the matrix.
    dsizes : ndarray
        An array contains ``dsize`` for each block.
    dsize : int
        Overall diagonal size of the matrix.
    lranks : int
        Ranks of each l-matrix block.
    sdiags : ndarray
        Square root of the diagonal.
    bilmat : BILMat
        Block ILMat for easier computation

    Methods
    -------
    dot(x)
        Dot product with vector or matrix.
    invdot(x)
        Inverse dot product with vector or matrix.
    logdet()
        Log determinant of the matrix.
    """

    def __init__(self, diags: Iterable, lmats: Iterable, dsizes: Iterable):
        self.diags = np.ascontiguousarray(diags)
        self.lmats = np.ascontiguousarray(lmats)
        self.dsizes = np.ascontiguousarray(dsizes)
        self.lranks = np.minimum(self.dsizes, self.lmats.shape[1])
        self.sdiags = np.sqrt(self.diags)

        self.bilmat = BILMat(self.lmats/self.sdiags[:, np.newaxis], self.dsizes)
        self.dsize = self.dsizes.sum()

    @property
    def mat(self) -> ndarray:
        return self.bilmat.mat * (self.sdiags[:, np.newaxis] * self.sdiags)

    @property
    def invmat(self) -> ndarray:
        return self.bilmat.invmat / (self.sdiags[:, np.newaxis] * self.sdiags)

    @property
    def num_blocks(self) -> int:
        return self.dsizes.size

    def dot(self, x: ndarray) -> ndarray:
        """
        Inverse dot product with vector or matrix

        Parameters
        ----------
        x : Iterable
            Vector or matrix

        Returns
        -------
        ndarray
        """
        x = np.ascontiguousarray(x)
        x = (x.T*self.sdiags).T
        x = self.bilmat.dot(x)
        x = (x.T*self.sdiags).T
        return x

    def invdot(self, x: ndarray) -> ndarray:
        """
        Inverse dot product with vector or matrix

        Parameters
        ----------
        x : Iterable
            Vector or matrix

        Returns
        -------
        ndarray
        """
        x = np.ascontiguousarray(x)
        x = (x.T/self.sdiags).T
        x = self.bilmat.invdot(x)
        x = (x.T/self.sdiags).T
        return x

    def logdet(self) -> float:
        """
        Log determinant

        Returns
        -------
        float
            Log determinant of the matrix.
        """
        return np.log(self.diags).sum() + self.bilmat.logdet()

    def __repr__(self) -> str:
        return f"BDLMat(dsize={self.dsize}, num_blocks={self.num_blocks})"



__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "spmat"
__summary__ = "spmat: A collection of tools for special matrices"
__uri__ = "https://github.com/zhengp0/spmat"

__version__ = "0.0.7"

__author__ = "Peng Zheng"
__email__ = "zhengp@uw.edu"

__license__ = "BSD 2-Clause License"
__copyright__ = f"Copyright 2021 {__author__}"



"""
Utility functions
"""
from typing import Iterable, Tuple, List
import numpy as np


def to_numpy(array: Iterable,
             ndim: Tuple[int] = None) -> np.ndarray:
    if not isinstance(array, np.ndarray):
        array = np.asarray(array)

    if ndim is not None:
        if array.ndim not in ndim:
            raise ValueError(f"`array` ndim must be in {ndim}.")

    return array


def split(array: Iterable,
          sizes: Iterable[int],
          axis: int = 0) -> List[np.ndarray]:
    array = to_numpy(array)
    if array.shape[axis] != sum(sizes):
        raise ValueError("`array` not match `sizes`.")
    return np.split(array, np.cumsum(sizes)[:-1], axis=axis)


def create_bdiag_mat(mats: List[np.ndarray]) -> np.ndarray:
    if not all([mat.ndim == 2 for mat in mats]):
        raise ValueError("`mats` must be a list of matrices.")
    row_sizes = [mat.shape[0] for mat in mats]
    col_sizes = [mat.shape[1] for mat in mats]
    row_size = sum(row_sizes)
    col_size = sum(col_sizes)

    bdiag_mat = np.zeros((row_size, col_size), dtype=mats[0].dtype)
    row_idx = split(np.arange(row_size), row_sizes)
    col_idx = split(np.arange(col_size), col_sizes)

    for i, mat in enumerate(mats):
        bdiag_mat[np.ix_(row_idx[i], col_idx[i])] = mat

    return bdiag_mat



import sys
from pathlib import Path

import numpy
from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup

if __name__ == "__main__":
    base_dir = Path(__file__).parent
    src_dir = base_dir / "src" / "spmat"

    sys.path.insert(0, src_dir.as_posix())
    import __about__ as about

    with (base_dir / "README.rst").open() as f:
        long_description = f.read()

    install_requirements = [
        "numpy",
        "scipy"
    ]

    test_requirements = [
        "pytest",
        "pytest-mock",
    ]

    doc_requirements = []

    setup(
        name=about.__title__,
        version=about.__version__,

        description=about.__summary__,
        long_description=long_description,
        license=about.__license__,
        url=about.__uri__,

        author=about.__author__,
        author_email=about.__email__,

        package_dir={"": "src"},
        packages=find_packages(where="src"),
        include_package_data=True,

        install_requires=install_requirements,
        tests_require=test_requirements,
        extras_require={
            "docs": doc_requirements,
            "test": test_requirements,
            "dev": doc_requirements + test_requirements
        },
        ext_modules=cythonize("src/spmat/linalg.pyx"),
        include_dirs=[numpy.get_include()],
        zip_safe=False,
    )



import numpy as np
cimport numpy as cnp
cimport cython
cimport scipy.linalg.cython_blas as blas
cimport scipy.linalg.cython_lapack as lapack


@cython.boundscheck(False)
@cython.wraparound(False)
def block_lsvd(double[:, ::1] a_view, long[::1] n_view, long[::1] k_view):
    cdef int dim_row = max(n_view)
    cdef int dim_col = a_view.shape[1]
    cdef int dim_max = max(dim_row, dim_col)
    cdef int dim_min = min(dim_row, dim_col)
    cdef int lwork = 2*max(1, 3*dim_min + dim_max, 5*dim_min)
    cdef int ind_a = 0
    cdef int ind_u = 0
    cdef int ind_s = 0
    cdef int info

    u = np.empty(np.array(n_view).dot(np.array(k_view)), dtype=np.float_)
    s = np.empty(np.array(k_view).sum(), dtype=np.float_)
    w = np.empty(lwork, dtype=np.float_)
    cdef double[::1] u_view = u
    cdef double[::1] s_view = s
    cdef double[::1] w_view = w


    for i in range(n_view.size):
        dim_row = n_view[i]
        dim_min = k_view[i]

        lapack.dgesvd("N", "S",
                      &dim_col, &dim_row,
                      &a_view[ind_a][0], &dim_col,
                      &s_view[ind_s],
                      &a_view[ind_a][0], &dim_col,
                      &u_view[ind_u], &dim_min,
                      &w_view[0], &lwork, &info)

        ind_a += dim_row
        ind_u += dim_row*dim_min
        ind_s += dim_min

    return u, s


@cython.boundscheck(False)
@cython.wraparound(False)
def block_mvdot(double[::1] u_view,
                double[::1] v_view,
                double[::1] x_view,
                long[::1] n_view,
                long[::1] k_view):
    cdef int dim_row
    cdef int dim_col
    cdef int one_int = 1
    cdef double zero_double = 0
    cdef double one_double = 1

    t = np.empty(np.array(k_view).sum(), dtype=np.float_)
    y = np.empty(np.array(n_view).sum(), dtype=np.float_)
    cdef double[::1] y_view = y
    cdef double[::1] t_view = t
    cdef int i
    cdef int ind_x = 0
    cdef int ind_y = 0
    cdef int ind_t = 0
    cdef int ind_u = 0

    # compute t = u.T @ x
    for i in range(n_view.size):
        dim_row = n_view[i]
        dim_col = k_view[i]

        blas.dgemv("N",
                   &dim_col, &dim_row, &one_double,
                   &u_view[ind_u], &dim_col,
                   &x_view[ind_x], &one_int, &zero_double,
                   &t_view[ind_t], &one_int)

        ind_x += dim_row
        ind_t += dim_col
        ind_u += dim_row*dim_col

    # compute t = t * v
    for i in range(t_view.size):
        t_view[i] *= v_view[i]

    # compute y = u @ t
    ind_t = 0
    ind_u = 0
    for i in range(n_view.size):
        dim_row = n_view[i]
        dim_col = k_view[i]

        blas.dgemv("T",
                   &dim_col, &dim_row, &one_double,
                   &u_view[ind_u], &dim_col,
                   &t_view[ind_t], &one_int, &zero_double,
                   &y_view[ind_y], &one_int)

        ind_y += dim_row
        ind_t += dim_col
        ind_u += dim_row*dim_col

    # compute y = y + x
    for i in range(y_view.size):
        y_view[i] += x_view[i]

    return y


@cython.boundscheck(False)
@cython.wraparound(False)
def block_mmdot(double[::1] u_view,
                double[::1] v_view,
                double[:, ::1] x_view,
                long[::1] n_view,
                long[::1] k_view):
    cdef int dim_row
    cdef int dim_col
    cdef int num_col = x_view.shape[1]
    cdef double zero_double = 0
    cdef double one_double = 1

    t = np.empty((np.array(k_view).sum(), num_col), dtype=np.float_)
    y = np.empty((np.array(n_view).sum(), num_col), dtype=np.float_)
    cdef double[:, ::1] y_view = y
    cdef double[:, ::1] t_view = t
    cdef int i, j
    cdef int ind_x = 0
    cdef int ind_y = 0
    cdef int ind_t = 0
    cdef int ind_u = 0

    # compute t = u.T @ x
    for i in range(n_view.size):
        dim_row = n_view[i]
        dim_col = k_view[i]
        
        blas.dgemm("N", "T",
                   &num_col, &dim_col, &dim_row, &one_double,
                   &x_view[ind_x][0], &num_col,
                   &u_view[ind_u], &dim_col, &zero_double,
                   &t_view[ind_t][0], &num_col)

        ind_x += dim_row
        ind_t += dim_col
        ind_u += dim_row*dim_col

    # compute t = t * v
    for i in range(t_view.shape[0]):
        for j in range(t_view.shape[1]):
            t_view[i, j] *= v_view[i]

    # compute y = u @ t
    ind_u = 0
    ind_t = 0
    for i in range(n_view.size):
        dim_row = n_view[i]
        dim_col = k_view[i]

        blas.dgemm("N", "N",
                   &num_col, &dim_row, &dim_col, &one_double,
                   &t_view[ind_t][0], &num_col,
                   &u_view[ind_u], &dim_col, &zero_double,
                   &y_view[ind_y][0], &num_col)

        ind_y += dim_row
        ind_t += dim_col
        ind_u += dim_row*dim_col

    # compute y = y + x
    for i in range(y_view.shape[0]):
        for j in range(y_view.shape[1]):
            y_view[i, j] += x_view[i, j]

    return y
